import pygame, math, random
from pygame import Rect
from pygame.sprite import DirtySprite, LayeredDirty, Sprite
import pygame.transform as transform
from util import loadimg
from vec2d import Vec2d
import gamestates
import spritesheets
from sprites import Animated, AnimationMode, NPCState
import controls, sound
from physics import PhysicsSprite, blit_collision_rect

class FurnitureSprite(PhysicsSprite):
    def __init__(self,*args,**kwargs):
        super(FurnitureSprite, self).__init__(*args, **kwargs)
        self._layer = None

        # Position of image within file
        self.offset = Vec2d(0,0)

        self.interactable = True

    def initImage(self, filename, x, y):
        self.image = loadimg(filename)
        self.rect = self.image.get_rect()
        self.setPos(x,y)

    def setPos(self,x,y):
        """Translate from screen coordinates to world coordinates."""
        self.x = x + self.rect.width/2 - self.offset.x

        # Bottom of image
        myY = y + self.rect.height - self.offset.y
        if hasattr(self, 'sprite_bottom_padding'):
            myY -= self.sprite_bottom_padding

        if myY < self.arena.top:
            self.y = myY - self.arena.top
            self.z = 0
        else:
            self.y = 0
            self.z = myY - self.arena.top
    
    def collide(self, sprite):
        #print 'Collide', self, 'with', sprite
        pass

class FurnitureSpriteAnimated(Animated, FurnitureSprite):
    def __init__(self,*args,**kwargs):
        super(FurnitureSpriteAnimated, self).__init__(*args, **kwargs)
        self.collision_rect_color = (0, 0xcc, 0, 0x80)

        self._show_collision_rect = 0

    @property
    def show_collision_rect(self):
        return self._show_collision_rect
    @show_collision_rect.setter
    def show_collision_rect(self,val):
        if self._show_collision_rect != val:
            self.dirty = 1
            self.force_frame_update = 1
        self._show_collision_rect = val

    def initImage(self, filename, x, y):
        self.sheet = spritesheets.sheets[filename]
        self.rect = self.sheet[0,0].get_rect()
        self.setPos(x,y)

    def transform_frame(self, image):
        if controls.show_collision_rects() or self.show_collision_rect:
            image = blit_collision_rect(self,image)
        return image


class BillboardSprite(FurnitureSpriteAnimated):
    def __init__(self,x,y,*args,**kwargs):
        super(BillboardSprite, self).__init__(*args, **kwargs)
        self.initImage('billboard_ss.png',x,y)
        self.animmode = AnimationMode.STOPPED
        self.sprite_bottom_padding = 46
        self.upside_down = 0
        self.z_width = 46
        self.counter = 0

    def interact(self,x,y,z):
        self.animmode = AnimationMode.FORWARD
        self.phase_through_platforms = True
        self.y += 10
        self.z += 10
        self.upside_down = 1
        self.interactable = False

    def stop_movement(self):
        super(BillboardSprite,self).stop_movement()
        sound.sounds['crash.wav'].play()

    def collide(self, sprite):
        super(BillboardSprite,self).collide(sprite)
        if self.vel.y != 0:
            sprite.damage(100)
            if self.counter>0: sprite.combo_kill()
            self.arena.game.score += self.counter
            self.counter += 1


class BirdSprite(FurnitureSpriteAnimated):
    def __init__(self,x,y,*args,**kwargs):
        super(BirdSprite, self).__init__(*args, **kwargs)
        self.initImage('bird.png',x,y)
        self.obey_gravity = False
        self.is_background = 1
        self.interactable = False
        self.mask_fudge_factor = 0


class CarSprite(FurnitureSprite):
    def __init__(self,x,y,*args,**kwargs):
        super(CarSprite, self).__init__(*args, **kwargs)
        self.initImage('car_ss.png',x,y)
        self.vel.x = -4
        self.bounded_in_x = 0
        self.simulate_during_pause = 0
        self.counter = 0

    def collide(self, sprite):
        super(CarSprite,self).collide(sprite)
        sprite.damage(100)
        if self.counter>0: sprite.combo_kill()
        self.arena.game.score += self.counter
        self.counter += 1

    def update(self,*args):
        super(CarSprite,self).update(*args)
        if self.rect.right < self.arena.x[0]:
            self.rect.left = self.arena.x[1]
            self.counter = 0
        elif self.rect.left > self.arena.x[1]:
            self.rect.right = self.arena.x[0]
            self.counter = 0

class CellarDoorSprite(FurnitureSpriteAnimated):
    def __init__(self,x,y,*args,**kwargs):
        super(CellarDoorSprite, self).__init__(*args, **kwargs)
        self.initImage('cellar_door_ss.png',x,y)
        self._layer = 0
        self.is_background = 1
        self.animmode = AnimationMode.STOPPED
        self.is_open = False
        self.counter = 0

    def interact(self,x,y,z):
        self.animmode = AnimationMode.FORWARD
        self.is_open = True
        self.interactable = False

    def collide(self, sprite):
        super(CellarDoorSprite,self).collide(sprite)
        # Make sure the *feet* of the character hit the trapdoor
        if self.is_open and sprite.isAlive() and self.rect.collidepoint(sprite.rect.midbottom):
            sprite.fall()
            if self.counter>0: sprite.combo_kill()
            self.arena.game.score += self.counter
            self.counter += 0.5

class FlowersSprite(FurnitureSpriteAnimated):
    def __init__(self,x,y,*args,**kwargs):
        self.sprite_bottom_padding = 15
        super(FlowersSprite, self).__init__(*args, **kwargs)
        self.offset = Vec2d(3,3)
        self.initImage('flowers_ss.png',x,y)
        self.animmode = AnimationMode.STOPPED
        self.sprite_bottom_padding = 15
        self.upside_down = 0
        self.z_width = 15

    def interact(self,x,y,z):
        self.animmode = AnimationMode.FORWARD
        self.phase_through_platforms = True
        self.y += 5
        self.z += 5
        self.upside_down = 1
        self.interactable = False

    def stop_movement(self):
        super(FlowersSprite,self).stop_movement()
        sound.sounds['crash.wav'].play()

    def collide(self, sprite):
        super(FlowersSprite,self).collide(sprite)
        if self.vel.y != 0:
            sprite.damage(100)


class ExplosionSprite(FurnitureSpriteAnimated):
    def __init__(self,x,y,*args,**kwargs):
        super(ExplosionSprite,self).__init__(*args,**kwargs)
        self.initImage('explosion.png',x,y)
        self.animmode = AnimationMode.FORWARD
        self.counter = 0

    def update(self,*args):
        super(ExplosionSprite,self).update(*args)
        if self.animmode != AnimationMode.STOPPED:
            mypos = Vec2d(self.x, self.z)
            for npc in self.arena.game.npcs:
                theirpos = Vec2d(npc.x, npc.z)
                if npc.isAlive() and (mypos-theirpos).length < self.rect.height/2:
                    npc.damage(1000)
                    if self.counter>0: npc.combo_kill()
                    self.arena.game.score += self.counter
                    self.counter += 1

class PropaneTankSprite(FurnitureSpriteAnimated):
    def __init__(self,x,y,*args,**kwargs):
        self.screen_pos = (x,y)
        super(PropaneTankSprite, self).__init__(*args, **kwargs)
        self.offset = Vec2d(4,0)
        self.initImage('propane_tank_ss.png',x,y)
        self.animmode = AnimationMode.STOPPED

    def interact(self,x,y,z):
        self.animmode = AnimationMode.FORWARD
        # TODO: kill nearby NPCs
        self.interactable = False

        size = spritesheets.sheets['explosion.png'][0,0].get_rect().size
        x = self.screen_pos[0] + self.rect.w/2 - size[0]/2
        y = self.screen_pos[1] + self.rect.h/2 - size[1]/2
        game = self.arena.game

        expl = ExplosionSprite(x,y,arena=self.arena)
        game.sprites.add(expl)
        game.furniture.add(expl)

        sound.sounds['explosion.wav'].play()

class ShopSignSprite(FurnitureSpriteAnimated):
    def __init__(self,x,y,*args,**kwargs):
        super(ShopSignSprite, self).__init__(*args, **kwargs)
        self.offset = Vec2d(8,0)
        self.initImage('shop_sign_ss.png',x,y)
        self.animmode = AnimationMode.STOPPED
        self.obey_gravity = False

    def interact(self,x,y,z):
        self.animmode = AnimationMode.FORWARD_RESET
        # TODO: kill NPC if collision

class StreetlightSprite(FurnitureSpriteAnimated):
    def __init__(self,x,y,*args,**kwargs):
        super(StreetlightSprite, self).__init__(*args, **kwargs)
        self.initImage('streetlight_ss.png',x,y)
        self.animmode = AnimationMode.STOPPED

    @property
    def collision_rect(self):
        r = Rect(0,0,15,40)
        r.bottomleft = self.rect.bottomleft
        return r

    #def interact(self,x,y,z):
    #    self.animmode = AnimationMode.FORWARD
    #    # TODO: kill NPC if collision

class TrashcanSprite(FurnitureSpriteAnimated):
    def __init__(self,x,y,*args,**kwargs):
        super(TrashcanSprite, self).__init__(*args, **kwargs)
        self.offset = Vec2d(2,0)
        self.initImage('trashcan_ss.png',x,y)
        self.animmode = AnimationMode.STOPPED
        self.phase_through_platforms = 1
        self.last_sprite_collided = None
        self.keepy_uppy = 0

    def interact(self,x,y,z):
        self.animmode = AnimationMode.LOOPED

        # Figure out which way the can should roll by comparing centres
        if x < self.rect.x+self.rect.width/2:
            self.vel.x = 3
        else:
            self.vel.x = -3

        self.simulate_during_pause = 1

    def stop_movement(self):
        self.vel.y = 0
        self.vel.x /= 2
        self.keepy_uppy = 0
        if abs(self.vel.x) < 1:
            self.animmode = AnimationMode.STOPPED

    def collide(self, sprite):
        super(TrashcanSprite,self).collide(sprite)
        if self.vel != (0,0) or self.vel_z != 0:
            if sprite == self.last_sprite_collided: return
            self.last_sprite_collided = sprite
            sprite.damage(10)
            if self.keepy_uppy>0: sprite.combo_kill()
            self.vel.x *= -1
            self.vel.y = -2*abs(self.vel.x)
            self.arena.game.score += self.keepy_uppy
            self.keepy_uppy += 1
            self.simulate_during_pause = 0
            sound.sounds['bounce.wav'].play()

class TreeSprite(FurnitureSprite):
    def __init__(self,x,y,*args,**kwargs):
        super(TreeSprite, self).__init__(*args, **kwargs)
        self.initImage('tree_ss.png',x,y)

WHEELIE_KICK_JUMP_VEL = -4

class WheelieBinSprite(FurnitureSprite):
    def __init__(self,x,y,*args,**kwargs):
        super(WheelieBinSprite, self).__init__(*args, **kwargs)
        self.initImage('wheelie_bin_ss.png',x,y)

    def interact(self,x,y,z):
        self.vel.y = WHEELIE_KICK_JUMP_VEL
        dx = self.x - x
        dz = self.z - z
        mag = math.sqrt(dx*dx + dz*dz)
        if mag == 0: mag = 1
        dx *= 2/mag
        dz *= 2/mag
        self.vel.x = dx
        self.vel_z = dz


class TimeLeftSprite(FurnitureSpriteAnimated):
    def __init__(self, game, *args,**kwargs):
        self.seconds = 10.000
        super(TimeLeftSprite,self).__init__(*args,**kwargs)
        self.game = game
        self.image = pygame.Surface((200,25))
        self.rect = self.image.get_rect()
        (self.x,self.y) = self.game.screen.get_rect().topright
        self.rect.topright = (self.x-5,self.y+5)
        self.update_meter()
        self.setPos(*self.rect.topleft)
        self.obey_gravity = False
        self.z_width = 50
        self.fallen = False
        self.counter = 0

    @property
    def image(self):
        if self.angle == 0:
            return self._image
        else:
            self._rotated = pygame.transform.rotate(self._image, self.angle)
            self._rotated = self.transform_frame(self._rotated)
            self._rotated_rect = self._rotated.get_rect()
            self._rotated_rect.x = self._rect.x
            self._rotated_rect.y = self._rect.y
            self._rotated_rect.midtop = self._rect.midtop
            return self._rotated
    @image.setter
    def image(self,val):
        self._image = val

    @property
    def rect(self):
        if self.angle == 0 or not hasattr(self, '_rotated_rect'):
            return self._rect
        else:
            return self._rotated_rect
    @rect.setter
    def rect(self,val):
        self._rect = val

    def update_meter(self):
        image = pygame.Surface((200,25), pygame.SRCALPHA, 32)
        image = image.convert_alpha()

        r,g = None, None
        # Fade from green to yellow in first 5 seconds
        # Fade from yellow to red in final 5 seconds
        if self.seconds > 5:
            r = int(255*((10-self.seconds)/5))
            g = 255
        else:
            r = 255
            g = int(255*(self.seconds/5))

        colour = pygame.Color(r,g,0,255)

        # Hacky access to _rect below
        length = self._rect.width*(self.seconds/10)
        rect = pygame.Rect(self._rect.width-length,0,length,self._rect.height)
        image.fill(colour, rect)
        pygame.draw.rect(image, (0,0,0), (0,0,self._rect.width,self._rect.height), 3)
        self.image = image        

    def update(self,dt,*args):
        super(TimeLeftSprite,self).update(dt,*args)
        if self.game.arena.time_paused:
            self.seconds -= dt
            if self.seconds <= 0:
                self.seconds = 0.0
                self.game.next_play_update = lambda: self.game.set_state(gamestates.PlayState(self.game))
        else:
            if self.seconds < 10:
                self.seconds += 4*dt
            if self.seconds > 10:
                self.seconds = 10
        self.update_meter()

        if self.vel.y != 0:
            for npc in self.arena.game.npcs:
                if npc.state != NPCState.DEAD and self.rect.left<npc.x<self.rect.right and self.z-self.z_width<npc.z<self.z+self.z_width:
                    npc.damage(200)
                    if self.counter>0: npc.combo_kill()
                    self.arena.game.score += self.counter
                    self.counter += 1

    def interact(self,x,y,z):
        if not self.fallen:
            self.phase_through_platforms = True
            self.obey_gravity = True
            self.y -= 50
            self.z += 50
            self.fallen = True
            self.rotation = 5
        self.interactable = False

    def collide(self, sprite):
        super(TimeLeftSprite,self).collide(sprite)
        if self.vel.y != 0:
            sprite.damage(100)
