import pygame
from pygame import Rect
from pygame.sprite import LayeredDirty, Group
from util import QuitException
from furniture import *
from sprites import *
from physics import sprite_collide_rect


class GameState(object):
    def __init__(self,game,*args,**kwargs):
        super(GameState,self).__init__(*args,**kwargs)
        self.game = game

    def update(self,*args):
        pass
    def render(self,screen):
        pass
    def event(self,ev):
        pass

    def enter(self):
        pass
    def leave(self):
        pass


class TitleState(GameState):
    def __init__(self,*args,**kwargs):
        super(TitleState,self).__init__(*args,**kwargs)
        self.arena = Arena(self.game, self.game.screen, None)
        self.arena.gravity = 0
        self.sprites = LayeredDirty()
        self.sprites.add(Background('title.png'),layer=-1)

        sprite = AntichristSprite(self.game,arena=self.arena)
        sprite.speed = 0
        (sprite.x,sprite.z) = self.game.screen.get_rect().center

        self.sprites.add(sprite)

    def update(self,*args):
        self.sprites.update(*args)

    def render(self,screen):
        pygame.display.update(self.sprites.draw(screen))

    def leave(self):
        sound.play_main_music()

    def event(self,ev):
        if (ev.type == pygame.KEYUP or ev.type == pygame.MOUSEBUTTONUP or
                ev.type == pygame.JOYBUTTONUP):
            self.game.set_state(PlayState(self.game))


class PlayState(GameState):
    def update(self,*args):
        self.game.sprites.update(*args)

        collisions = groupcollide(self.game.furniture, self.game.npcs,
            False, False, sprite_collide_rect)
        for (furniture, npc_list) in collisions.iteritems():
            for npc in npc_list:
                if npc.state == NPCState.DEAD: continue
                z_width = 20
                if hasattr(furniture,'z_width'):
                    z_width = furniture.z_width
                if hasattr(furniture,'upside_down') and furniture.upside_down:
                    z_collide = furniture.z + z_width >= npc.z >= furniture.z
                else:
                    z_collide = furniture.z - z_width <= npc.z <= furniture.z
                if z_collide:
                    furniture.collide(npc)

        ac = self.game.antichrist
        if ac.state in (NPCState.DEAD, NPCState.FALLEN) and \
                ac.frameno == len(ac.frames)-1:
            self.game.set_state(GameOverState(self.game))

        if self.game.next_play_update:
            self.game.next_play_update()
            self.game.next_play_update = None
        
    def render(self,screen):
        rects = self.game.sprites.draw(screen)
        pygame.display.update(rects)

    def create_next_state(self):
        return PausedState(self.game)

    def event(self,ev):
        if controls.is_pause(ev):
            self.game.set_state(self.create_next_state())
        elif controls.is_show_collision_rects(ev):
            for spr in self.game.sprites.sprites():
                spr.dirty = 1
                if isinstance(spr,Animated):
                    spr.force_frame_update = 1


class PausedState(PlayState):
    def create_next_state(self):
        return PlayState(self.game)

    def event(self,ev):
        super(PausedState,self).event(ev)
        if controls.is_action(ev):
            if self.game.shadow is not None:
                def callback(furniture,sprite):
                    furniture.interact(sprite.x,sprite.y,sprite.z)
                self.game.shadow.search_for_interactions(callback)

    def enter(self):
        self.game.arena.time_paused = True
        self.game.show_shadow()
        self.game.controls_p.text = 'Z - Unpause'
        self.game.controls_j.visible = 1
        self.game.controls_j.dirty = 1
    def leave(self):
        self.game.arena.time_paused = False
        self.game.hide_shadow()
        self.game.controls_p.text = 'Z - Pause'
        self.game.controls_j.visible = 0
        self.game.controls_j.dirty = 1
        for o in self.game.sprites.sprites():
            o.show_collision_rect = 0
            o.dirty = 1


class GameOverState(GameState):
    def __init__(self,*args,**kwargs):
        super(GameOverState,self).__init__(*args,**kwargs)
        self.sprites = LayeredDirty()
        self.sprites.add(Background('screen_game_over.png'))
        self.score = TextSprite(str(int(self.game.score)))
        self.score.font = pygame.font.SysFont(pygame.font.get_default_font(),
                60)
        self.score.foreground = (0,0,0)
        self.score.rect.midleft = Vec2d(350, 336)
        self.sprites.add(self.score)

        self.first_render = 1

    def enter(self):
        sound.stop_music()
        sound.play_gameover_music()

    def leave(self):
        sound.stop_music()

    def update(self,*args):
        self.sprites.update(*args)

    def render(self,screen):
        rects = self.sprites.draw(screen)
        pygame.display.update(rects)

    def event(self,ev):
        if (ev.type == pygame.KEYUP or ev.type == pygame.MOUSEBUTTONUP or
                ev.type == pygame.JOYBUTTONUP):
            self.game.reset_game()


HUD_LAYER = 1000

class Game(object):
    def __init__(self,screen):
        self.screen = screen
        self.reset_game()

    def set_state(self, st):
        if self._state is not None:
            self._state.leave()
        self._state = st
        st.enter()

    def show_shadow(self):
        self.shadow = ShadowCharSprite(self,arena=self.arena)
        self.shadow.x = self.antichrist.x
        self.shadow.z = self.antichrist.z
        self.sprites.add(self.shadow)
    
    def hide_shadow(self):
        self.shadow.kill()
        self.shadow = None
        self.set_can_interact(False)

    def set_can_interact(self,val):
        # TODO:
        # * Update controls text
        if self._can_interact != val:
            self.exclamation_mark.set_visible(val)
            self.controls_a.visible = val
            self.controls_a.dirty = 1
        self._can_interact = val

    def reset_game(self):
        self.score = 0
        self._state = None
        self._can_interact = False
        self.next_play_update = None
        self.set_state(TitleState(self))

        arena = Arena(self, self.screen, Platforms('platforms.png'))
        self.arena = arena
        r = self.screen.get_rect()
        arena.x = (-32, r.width+64)
        arena.y = (-1000, 0)
        arena.z = (0, r.height+16-320)
        arena.top = 320

        self.sprites = LayeredDirty()

        self.npcs = Group()
        for i in range(15):
            sprite = NPCSprite(self, arena=arena)
            self.npcs.add(sprite)
        self.antichrist = AntichristSprite(self,arena=arena)
        self.npcs.add(self.antichrist)
        self.sprites.add(self.npcs.sprites())

        self.shadow = None # set only during pause

        self.sprites.add(Background('background.png'), layer=-1)
        self.time_left_sprite = TimeLeftSprite(self, arena=arena)
        self.sprites.add(self.time_left_sprite, layer=0)

        self.exclamation_mark = ExclamationMarkSprite(self)
        self.exclamation_mark.visible = 0
        self.sprites.add(self.exclamation_mark)

        self.controls_p = TextSprite('Z - Pause')
        self.controls_p.auto_adjust_pos = False
        self.controls_p.rect.center = self.time_left_sprite.rect.center

        self.controls_j = TextSprite('C - Jump')
        self.controls_j.visible = 0
        self.controls_j.rect.topleft = r.topleft

        self.controls_a = TextSprite('X - Action')
        self.controls_a.visible = 0
        self.controls_a.rect.topleft = self.controls_j.rect.bottomleft
        self.sprites.add(self.controls_p, self.controls_j,
            self.controls_a, layer=HUD_LAYER)

        furniture = self.furniture = Group()
        #furniture.add(AerialSprite(41,62, arena=arena))
        #furniture.add(AerialSprite(142,75, arena=arena))
        furniture.add(BillboardSprite(225,55, arena=arena))
        furniture.add(CarSprite(412,382, arena=arena))
        furniture.add(CellarDoorSprite(314,322, arena=arena))
        furniture.add(FlowersSprite(27,227, arena=arena))
        furniture.add(FlowersSprite(157,120, arena=arena))
        furniture.add(FlowersSprite(267,176, arena=arena))
        furniture.add(FlowersSprite(605,72, arena=arena))
        furniture.add(PropaneTankSprite(265,304, arena=arena))
        furniture.add(PropaneTankSprite(560,394, arena=arena))
        furniture.add(ShopSignSprite(74,268, arena=arena))
        furniture.add(ShopSignSprite(303,268, arena=arena))
        furniture.add(StreetlightSprite(24,281, arena=arena))
        furniture.add(StreetlightSprite(190,281, arena=arena))
        furniture.add(StreetlightSprite(387,281, arena=arena))
        furniture.add(StreetlightSprite(553,281, arena=arena))
        furniture.add(TrashcanSprite(325,389, arena=arena))
        furniture.add(TrashcanSprite(140,321, arena=arena))
        #furniture.add(TreeSprite(425,304, arena=arena))
        furniture.add(WheelieBinSprite(124,297, arena=arena))
        furniture.add(WheelieBinSprite(379,295, arena=arena))
        furniture.add(WheelieBinSprite(531,298, arena=arena))
        furniture.add(self.time_left_sprite)
        self.sprites.add(furniture.sprites())

        background = self.background = Group()
        self.background.add(BirdSprite(380,30, arena=arena))
        self.background.add(BirdSprite(360,24, arena=arena))
        self.background.add(BirdSprite(372,28, arena=arena))
        self.background.add(BirdSprite(392,26, arena=arena))
        self.sprites.add(background.sprites())
    
    def update(self, *args):
        self._state.update(*args)

        # Hacky hack to make sure layers are updated
        for s in self.sprites.sprites():
            self.sprites.change_layer(s, s._layer)

    def render(self,screen):
        self._state.render(screen)

    def event(self,ev):
        #print ev
        if ev.type == pygame.QUIT or (ev.type == pygame.KEYUP and
                ev.key == pygame.K_ESCAPE):
            raise QuitException()
        self._state.event(ev)

