import pygame
from pygame import Rect
from pygame.sprite import DirtySprite
from vec2d import Vec2d
from util import loadimg


DEFAULT_GRAVITY = 0.5


class PhysicsSprite(DirtySprite):
    def __init__(self,*args,**kwargs):
        assert 'arena' in kwargs
        self.arena = kwargs.pop('arena')
        super(PhysicsSprite,self).__init__(*args,**kwargs)

        self._y = 0
        self._z = 0
        self.vel = Vec2d(0,0)
        self.vel_z = 0
        self.angle = 0
        self.rotation = 0

        self.is_background = 0
        self.phase_through_platforms = 0
        self.sprite_bottom_padding = 0
        self.obey_gravity = 1
        self.bounded_in_x = 1
        self.simulate_during_pause = 1
        self.mask_fudge_factor = 6

        self.collision_rect_color = (0xff, 0, 0, 0x80)

    @property
    def collision_rect(self):
        r = Rect(self.rect)
        r.height -= self.sprite_bottom_padding
        if hasattr(self,'upside_down') and self.upside_down:
            r.bottom = self.rect.bottom
        return r

    def drawable_collision_rect(self):
        r = self.collision_rect
        img = pygame.Surface(r.size).convert_alpha()
        img.fill(self.collision_rect_color)
        return img

    @property
    def x(self): return self.rect.centerx
    @x.setter
    def x(self,val): self._rect.centerx = val

    @property
    def y(self): return self._y
    @y.setter
    def y(self,val):
        self._y = val
        self._update_yz()

    def _update_yz(self):
        self._rect.bottom = self.y + self.z + self.arena.top + self.sprite_bottom_padding
        if hasattr(self, '_rotated_rect'):
            self._rotated_rect.bottom = self._rect.bottom

    @property
    def z(self): return self._z
    @z.setter
    def z(self,val):
        self._z = val
        if self.is_background:
            self._layer = 0
        else:
            self._layer = val
        self._update_yz()

    @property
    def rect(self):
        return self._rect
    @rect.setter
    def rect(self,val):
        self._rect = val

    def standing_on_platform(self):
        if self.vel.y < 0: return False
        if self.phase_through_platforms: return False
        if self.arena.platforms is None: return False
        offset = Vec2d(self.rect.x+3, self.rect.bottom-5-self.sprite_bottom_padding)
        selfmask = pygame.mask.Mask((self.rect.width-self.mask_fudge_factor, self.mask_fudge_factor))
        selfmask.fill()
        return self.arena.platforms.collide(selfmask, offset)

    def grounded(self):
        '''
        Is the player on the ground or on a platform?
        '''
        return self.y == self.arena.y[1] or self.standing_on_platform()

    def stop_movement(self):
        self.vel = Vec2d(0,0)
        self.vel_z = 0

    def update(self,*args):
        super(PhysicsSprite,self).update(*args)
        if self.arena.time_paused and not self.simulate_during_pause:
            return
        orig_vel_y = self.vel.y

        self.x += self.vel.x
        self.y += self.vel.y
        self.z += self.vel_z
        if self.vel != (0,0) or self.vel_z != 0:
            self.dirty = 1

        if self.obey_gravity:
            self.vel.y += self.arena.gravity

        if self.y != self.arena.y[1]:   # only rotate when off the ground
            self.angle += self.rotation
        else:
            self.rotation = 0
            # snap angle to nearest 90 degrees?

        on_plat = self.standing_on_platform()
        if self.y >= self.arena.y[1] or on_plat:
            if not on_plat:
                self.y = self.arena.y[1]
            if self.vel.y > 0:
                if orig_vel_y:
                    # if we were moving vertically, stop all movement when
                    # we hit the ground
                    self.stop_movement()
                else:
                    # otherwise if we were only moving horizontally, keep
                    # moving that way
                    self.vel.y = 0
        elif self.y <= self.arena.y[0]:
            self.y = self.arena.y[0]
            if self.vel.y < 0:
                self.vel.y = 0
        if self.bounded_in_x:
            if self.x >= self.arena.x[1]:
                self.x = self.arena.x[1]
                if self.vel.x > 0:
                    self.vel.x = 0
            elif self.x <= self.arena.x[0]:
                self.x = self.arena.x[0]
                if self.vel.x < 0:
                    self.vel.x = 0
        if self.z >= self.arena.z[1]:
            self.z = self.arena.z[1]
            if self.vel_z > 0:
                self.vel_z *= -1
        elif self.z <= self.arena.z[0]:
            self.z = self.arena.z[0]
            if self.vel_z < 0:
                self.vel_z *= -1


class Arena(object):
    def __init__(self, game, screen, platforms):
        self.game = game
        r = screen.get_rect()
        self.screen_r = r
        self.x = (0,r.width)
        self.y = (0,r.height)
        self.z = (0,r.height)
        self.top = 0
        self.platforms = platforms
        self.gravity = DEFAULT_GRAVITY
        self.time_paused = False


class Platforms(object):
    def __init__(self, fname):
        self.mask = pygame.mask.from_surface(loadimg(fname))

    def collide(self, othermask, offset):
        return self.mask.overlap(othermask, offset) is not None


def get_collision_rect(sprite):
    if hasattr(sprite, 'collision_rect'):
        rect = sprite.collision_rect
    else:
        rect = sprite.rect
    return rect
    
def sprite_collide_rect(spr1, spr2):
    r1, r2 = (get_collision_rect(spr1),
              get_collision_rect(spr2))
    return r1.colliderect(r2)
    

def blit_collision_rect(sprite, image):
    newimage = pygame.Surface(image.get_rect().size).convert_alpha()
    newimage.fill((0,0,0,0))
    cr = sprite.collision_rect
    tl = Vec2d(cr.topleft)
    newimage.blit(sprite.drawable_collision_rect(), tl - sprite.rect.topleft)
    newimage.blit(image, (0,0))
    return newimage
