import time, math
from random import choice, randint
import pygame
import pygame.draw as draw
import pygame.transform as transform
from pygame.sprite import Sprite, groupcollide, Group
import spritesheets, controls, vec2d
from physics import *
import gamestates, sound
from util import enum
from pygame import Rect


AnimationMode = enum(
    LOOPED = 0,
    STOPPED = 1,
    FORWARD = 2,
    FORWARD_RESET = 3)

class Animated(Sprite):
    def __init__(self, *args,**kwargs):
        super(Animated,self).__init__(*args,**kwargs)
        # subclasses must define these attrs:
        self.sheet = None
        self._frames = None
        self.framerow = 0
        self.framerate = 10
        self.animmode = AnimationMode.LOOPED
        self.animtime = 0.0
        self.frameno = 0
        self.image = pygame.Surface((1,1))
        self.force_frame_update = 1

    @property
    def framerow(self):
        return self._framerow
    @framerow.setter
    def framerow(self,val):
        self._framerow = val
        self._frames = None

    @property
    def frames(self):
        if self._frames is not None:
            return self._frames
        flist = []
        for i in xrange(self.sheet.cols):
            flist.append((i,self.framerow))
        return flist

    def transform_frame(self,image):
        return image

    def update(self, dt, *args):
        super(Animated,self).update(*args)
        cur_frameno = self.frameno
        if self.animmode != AnimationMode.STOPPED and \
                self.sheet is not None:
            self.animtime += dt
            frame_period = 1.0/self.framerate
            while self.animtime > frame_period:
                self.animtime -= frame_period
                self.frameno += 1
            if self.frameno >= len(self.frames):
                if self.animmode == AnimationMode.LOOPED:
                    self.frameno %= len(self.frames)
                elif self.animmode == AnimationMode.FORWARD_RESET:
                    self.frameno = 0
                    self.animmode = AnimationMode.STOPPED
                else:
                    assert self.animmode == AnimationMode.FORWARD
                    self.frameno = len(self.frames)-1
                    self.animmode = AnimationMode.STOPPED
        if (self.frameno != cur_frameno or self.force_frame_update) and\
                self.sheet is not None:
            self.dirty = 1
            self.image = self.transform_frame(self.sheet[
                self.frames[self.frameno]])
        self.force_frame_update = 0


class TextSprite(DirtySprite):
    def __init__(self,text,*args,**kwargs):
        super(TextSprite,self).__init__(*args,**kwargs)
        self.font = pygame.font.SysFont(pygame.font.get_default_font(), 30)
        self._foreground = (0xff,0xff,0xff)
        self.x = 0
        self.y = 0
        self.auto_adjust_pos = True
        self.text = text

    @property
    def foreground(self):
        return self._foreground
    @foreground.setter
    def foreground(self,val):
        self._foreground = val
        self.refresh_text()

    @property
    def text(self): return self._text
    @text.setter
    def text(self,val):
        self._text = val
        self.refresh_text()

    def refresh_text(self):
        self.image = self.font.render(self._text, 1, self.foreground)
        if self.auto_adjust_pos:
            self.reset_pos()
        self.dirty = 1

    def reset_pos(self):
        self.rect = self.image.get_rect()
        self.rect.topleft = (self.x,self.y)


class CharacterSprite(Animated,PhysicsSprite):

    _DIRECTIONS = [-1,1] # -1 = left, 1 = right
    _RAILS = [20, 55, 90] # y positions NPCs can live in

    def __init__(self,filename,*args,**kwargs):
        super(CharacterSprite,self).__init__(*args,**kwargs)
        self.setSheet(filename, 1)
        self.direction = 1
        self.speed = 1
        self.force_frame_update = 0

        self._show_collision_rect = 0

    @property
    def show_collision_rect(self):
        return self._show_collision_rect
    @show_collision_rect.setter
    def show_collision_rect(self,val):
        if self._show_collision_rect != val:
            self.dirty = 1
            self.force_frame_update = 1
        self._show_collision_rect = val

    def setSheet(self, filename, scaleFactor):
        self.sheet = spritesheets.sheets[filename]
        self.rect = self.sheet[0,0].get_rect()
        self.rect.width *= scaleFactor
        self.rect.height *= scaleFactor

    def transform_frame(self,image):
        if self.direction == -1:
            image = transform.flip(image, 1, 0)

        image = transform.scale(image, (self.rect.width, self.rect.height))

        if controls.show_collision_rects() or self.show_collision_rect:
            image = blit_collision_rect(self,image)

        return image


NPCState = enum(
    HEALTHY = 0,
    INJURED = 1,
    FALLEN = 2,
    DEAD = 3,
)

NPC_KICK_JUMP_VEL = -4

class NPCSprite(CharacterSprite):

    def __init__(self,game,*args,**kwargs):
        if 'sheet' in kwargs:
            sheet = kwargs.pop('sheet')
        else:
            sheet = 'generic_character_ss.png'
        super(NPCSprite,self).__init__(sheet,*args,**kwargs)

        self.game = game

        # Initial position
        self.rail = choice(CharacterSprite._RAILS)
        self.x = randint(0,640) # FIXME: screen width hack
        self.sprite_bottom_padding = 2

        self.direction = choice(CharacterSprite._DIRECTIONS)

        self.hp = 20
        self.state = NPCState.HEALTHY
        self.interactable = True

    def interact(self,x,y,z):
        #print 'Interact', self, (x,y,z)
        self.vel.y = NPC_KICK_JUMP_VEL
        dx = self.x - x
        dz = self.z - z
        mag = math.sqrt(dx*dx + dz*dz)
        if mag == 0: mag = 1
        dx *= 2/mag
        dz *= 2/mag
        self.vel.x = dx
        self.vel_z = dz
        self.damage(2)

    def use_death_sheet(self):
        self.setSheet('generic_death_ss.png',1)

    def isAlive(self):
        return self.state == NPCState.HEALTHY or self.state == NPCState.INJURED

    def death_sound(self):
        sound.sounds['death.wav'].play()

    @property
    def state(self):
        return self._state
    @state.setter
    def state(self,val):
        self._state = val
        if val == NPCState.HEALTHY:
            self.speed = 1
            self.framerate = 10
            self.framerow = 0
        elif val == NPCState.INJURED:
            self.speed = 1
            self.framerate = 5
            self.framerow = 1
        elif val == NPCState.FALLEN:
            self.speed = 0
            self.framerow = 2
            self.animmode = AnimationMode.FORWARD
            self.interactible = False
            self.arena.game.score += 1
            self.death_sound()
        elif val == NPCState.DEAD:
            self.speed = 0
            self.framerate = 5
            self.framerow = 0
            x,y = self.rect.x, self.rect.y
            self.use_death_sheet()
            self.rect.x, self.rect.y = x,y
            self.animmode = AnimationMode.FORWARD 
            self.interactible = False
            self.arena.game.score += 1
            self.death_sound()
        self.frameno = 0
        self.dirty = 1

    def damage(self, dmg):
        if dmg <= 0: return
        sound.sounds['hurt.wav'].play()
        self.hp -= dmg
        if self.hp < 10 and self.state == NPCState.HEALTHY:
            self.state = NPCState.INJURED
        if self.hp <= 0 and self.state != NPCState.DEAD:
            self.state = NPCState.DEAD

    def fall(self):
        if self.state == NPCState.HEALTHY or self.state == NPCState.INJURED:
            self.state = NPCState.FALLEN

    def combo_kill(self):
        self.game.sprites.add(ComboKillSprite(self))

    def update(self,*args):
        super(CharacterSprite,self).update(*args)
        if self.arena.time_paused:
            self.animmode = AnimationMode.STOPPED
            return
        else:
            if self.state == NPCState.DEAD or self.state == NPCState.FALLEN:
                self.animmode = AnimationMode.FORWARD
            else:
                self.animmode = AnimationMode.LOOPED
        self.x += self.speed*self.direction

        if self.rect.left < self.arena.x[0] or self.rect.right > self.arena.x[1]:
            self.leftScreen()

        if self.state == NPCState.FALLEN and self.frameno == len(self.frames)-1:
            self.kill()
    
    def leftScreen(self):
        # World wraps around
        if self.rect.left < self.arena.x[0]:
            self.rect.right = self.arena.x[1]

        if self.rect.right > self.arena.x[1]:
            self.rect.left = self.arena.x[0]
        
        self.rail = choice(CharacterSprite._RAILS)

    def setRail(self, val):
        self._rail = val
        self.z = val

    @property
    def rail(self): return self._rail
    @rail.setter
    def rail(self,val):
        self._rail = val
        self.z = val


class AntichristSprite(NPCSprite):
    def __init__(self,game,*args,**kwargs):
        super(AntichristSprite,self).__init__(game,*args,
            sheet='antichrist_character_ss.png',**kwargs)
        self.setRail(1);
        self.x = 0
        self.z = CharacterSprite._RAILS[self._rail]
        self.direction = CharacterSprite._DIRECTIONS[1]
        self.game = game

    def death_sound(self):
        sound.sounds['WilhelmScream.wav'].play()

    def use_death_sheet(self):
        self.setSheet('antichrist_death_ss.png',1)

    def leftScreen(self):
        self.kill()

    def kill(self):
        super(AntichristSprite,self).kill()
        self.game.set_state(gamestates.GameOverState(self.game))


JUMP_VEL = -8

class ShadowCharSprite(CharacterSprite):
    def __init__(self,game,*args,**kwargs):
        super(ShadowCharSprite,self).__init__('shadow_character_ss.png',*args,**kwargs)

        self.x = 320
        self.z = choice(CharacterSprite._RAILS)
        self.sprite_bottom_padding = 2
        self.game = game
        self.can_interact = False

        self.speed = 3
        self.idle = 1

    @property
    def idle(self):
        return self._idle
    @idle.setter
    def idle(self, val):
        if val:
            self.image = self.transform_frame(self.sheet[0,1])
            self.animmode = AnimationMode.STOPPED
            self.dirty = 1
            self._idle = 1
        else:
            self.frameno = 0
            self.image = self.transform_frame(self.sheet[0,0])
            self.animmode = AnimationMode.LOOPED
            self.dirty = 1
            self._idle = 0            

    def search_for_interactions(self,callback):
        objects = [o for o in self.game.sprites if hasattr(o,'interact')]
        selfgroup = Group(self)
        objgroup = Group([o for o in objects if o.interactable])
        collisions = groupcollide(selfgroup, objgroup, False, False,
                sprite_collide_rect)
        if self in collisions:
            for o in collisions[self]:
                #print 'Collided with', coll
                callback(o, self)
            for o in objects:
                o.show_collision_rect = o in collisions[self]
        else:
            for o in objects:
                o.show_collision_rect = 0
        objgroup.empty()
        selfgroup.empty()

    def update(self,*args):
        super(ShadowCharSprite,self).update(*args)

        direction = controls.x()

        self.x += self.speed*direction
        self.phase_through_platforms = 0
        if not self.standing_on_platform() and self.grounded():
            self.z += self.speed*controls.y()
            if self.z < 0: self.z = 0
        else:
            if controls.y() > 0:
                self.phase_through_platforms = 1

        if direction < 0:
            self.direction = -1
        elif direction > 0:
            self.direction = 1

        stationary = controls.x() == 0 and controls.y() == 0 and not controls.jump()
        if not self.idle and stationary:
            self.idle = True
        elif self.idle and not stationary:
            self.idle = False

        if self.grounded() and controls.jump():
            self.vel.y = JUMP_VEL
            sound.sounds['jump.wav'].play()

        self.can_interact = False
        def callback(furniture,sprite):
            self.can_interact = True
        self.search_for_interactions(callback)
        self.game.set_can_interact(self.can_interact)


class ExclamationMarkSprite(TextSprite):
    def __init__(self,game,*args,**kwargs):
        super(ExclamationMarkSprite,self).__init__('!',*args,**kwargs)
        self._foreground = (0, 0xcc, 0)
        self.font = pygame.font.SysFont(pygame.font.get_default_font(), 50)
        self.game = game
        self.auto_adjust_pos = False
        self.refresh_text()

    def set_visible(self, val):
        self.visible = val
        self.dirty = 1

    def update(self,*args):
        super(ExclamationMarkSprite,self).update(*args)
        shadow = self.game.shadow
        if shadow is not None:
            (x,y) = shadow.rect.midtop
            (cx,cy) = self.rect.midbottom
            if cx != x or cy != y:
                self.rect.midbottom = (x,y)
                self.dirty = 1

class ComboKillSprite(Animated, DirtySprite):
    def __init__(self, sprite, *args, **kwargs):
        super(ComboKillSprite,self).__init__(*args, **kwargs)
        self.animmode = AnimationMode.FORWARD
        self.sheet = spritesheets.sheets['combo.png']
        self.rect = self.sheet[0,0].get_rect()
        self.rect.center = sprite.rect.center
        self._layer = sprite._layer


class Background(DirtySprite):
    def __init__(self,fname,*args,**kwargs):
        super(Background,self).__init__(*args,**kwargs)
        self.image = loadimg(fname)
        self.rect = self.image.get_rect()



